var Direction;
(function (Direction) {
    Direction[Direction["FORWARD"] = 0] = "FORWARD";
    Direction[Direction["BACKWARD"] = 1] = "BACKWARD";
})(Direction || (Direction = {}));
export var ValueType;
(function (ValueType) {
    ValueType[ValueType["REAL"] = 0] = "REAL";
    ValueType[ValueType["COMPLEX"] = 1] = "COMPLEX";
})(ValueType || (ValueType = {}));
class Buffer {
    type;
    array;
    _exports;
    constructor(arrayType, n, type, exports) {
        this.type = type;
        const arrayClass = arrayType;
        n *= type === ValueType.REAL ? 1 : 2;
        const alloc = arrayType === Float32Array
            ? exports.pffft_aligned_malloc
            : exports.pffftd_aligned_malloc;
        const ptr = alloc(arrayClass.BYTES_PER_ELEMENT * n);
        if (ptr === 0) {
            throw new Error("Failed to allocate memory");
        }
        this.array = new arrayClass(exports.memory.buffer, ptr, n);
        this._exports = exports;
    }
    [Symbol.dispose]() {
        const exports = this._exports;
        const free = this.ctor === Float32Array
            ? exports.pffft_aligned_free
            : exports.pffftd_aligned_free;
        free(this.array.byteOffset);
        this.array = null;
    }
    get ptr() {
        return this.array.byteOffset;
    }
    get _memory() {
        return this.array.buffer;
    }
    get ctor() {
        return this.array.constructor;
    }
}
class SetupImplFloat32 {
    #ptr;
    #exports;
    constructor(n, type, exports) {
        this.#ptr = exports.pffft_new_setup(n, type);
        if (this.#ptr === 0) {
            throw new Error("Failed to create PFFFT setup");
        }
        this.#exports = exports;
    }
    [Symbol.dispose]() {
        this.#exports.pffft_destroy_setup(this.#ptr);
        this.#ptr = 0;
    }
    transform(input, output, work, direction) {
        this.#exports.pffft_transform(this.#ptr, input, output, work, direction);
    }
    transformOrdered(input, output, work, direction) {
        this.#exports.pffft_transform_ordered(this.#ptr, input, output, work, direction);
    }
    zreorder(input, output, direction) {
        this.#exports.pffft_zreorder(this.#ptr, input, output, direction);
    }
    zconvolveAccu(dftA, dftB, dftAB, scaling) {
        this.#exports.pffft_zconvolve_accumulate(this.#ptr, dftA, dftB, dftAB, scaling);
    }
    zconvolveNoAccu(dftA, dftB, dftAB, scaling) {
        this.#exports.pffft_zconvolve_no_accu(this.#ptr, dftA, dftB, dftAB, scaling);
    }
}
class SetupImplFloat64 {
    #ptr;
    #exports;
    constructor(n, type, exports) {
        this.#ptr = exports.pffftd_new_setup(n, type);
        if (this.#ptr === 0) {
            throw new Error("Failed to create PFFFT setup");
        }
        this.#exports = exports;
    }
    [Symbol.dispose]() {
        this.#exports.pffftd_destroy_setup(this.#ptr);
        this.#ptr = 0;
    }
    transform(input, output, work, direction) {
        this.#exports.pffftd_transform(this.#ptr, input, output, work, direction);
    }
    transformOrdered(input, output, work, direction) {
        this.#exports.pffftd_transform_ordered(this.#ptr, input, output, work, direction);
    }
    zreorder(input, output, direction) {
        this.#exports.pffftd_zreorder(this.#ptr, input, output, direction);
    }
    zconvolveAccu(dftA, dftB, dftAB, scaling) {
        this.#exports.pffftd_zconvolve_accumulate(this.#ptr, dftA, dftB, dftAB, scaling);
    }
    zconvolveNoAccu(dftA, dftB, dftAB, scaling) {
        this.#exports.pffftd_zconvolve_no_accu(this.#ptr, dftA, dftB, dftAB, scaling);
    }
}
function mismatchedInstanceError() {
    return new Error("Mismatched PFFFT instances");
}
function checkFftBuffers(memory, work, signal, spectrum) {
    if (memory !== signal._memory || memory !== spectrum._memory) {
        throw mismatchedInstanceError();
    }
    const minLen = work.array.length;
    if (signal.array.length < minLen) {
        throw new Error("signal buffer too small");
    }
    if (spectrum.array.length < minLen) {
        throw new Error("spectrum buffer too small");
    }
}
function checkConvolveBuffers(memory, work, a, b, ab) {
    if (memory !== a._memory || memory !== b._memory || memory !== ab._memory) {
        throw mismatchedInstanceError();
    }
    const minLen = work.array.length;
    if (a.array.length < minLen) {
        throw new Error("a buffer too small");
    }
    if (b.array.length < minLen) {
        throw new Error("b buffer too small");
    }
    if (ab.array.length < minLen) {
        throw new Error("ab buffer too small");
    }
}
export class Setup {
    arrayType;
    n;
    work;
    #impl;
    #memory;
    constructor(arrayType, type, n, exports) {
        this.arrayType = arrayType;
        this.n = n;
        if (arrayType === Float32Array) {
            this.#impl = new SetupImplFloat32(n, type, exports);
        }
        else {
            this.#impl = new SetupImplFloat64(n, type, exports);
        }
        this.#memory = exports.memory.buffer;
        this.work = new Buffer(arrayType, n, type, exports);
    }
    [Symbol.dispose]() {
        this.#impl[Symbol.dispose]();
        this.work[Symbol.dispose]();
    }
    get type() {
        return this.work.type;
    }
    makeSignalBuffer(extra = 0) {
        return new Buffer(this.work.ctor, this.n + extra, this.work.type, this.work._exports);
    }
    makeSpectrumBuffer(extra = 0) {
        const n = (this.work.type === ValueType.COMPLEX ? this.n : this.n / 2) + extra;
        return new Buffer(this.work.ctor, n, ValueType.COMPLEX, this.work._exports);
    }
    makeInternalLayoutBuffer(extra = 0) {
        const n = (this.work.type === ValueType.COMPLEX ? 2 : 1) * (this.n + extra);
        return new Buffer(this.work.ctor, n, ValueType.REAL, this.work._exports);
    }
    forward(signal, spectrum) {
        const work = this.work;
        checkFftBuffers(this.#memory, work, signal, spectrum);
        this.#impl.transformOrdered(signal.ptr, spectrum.ptr, work.ptr, Direction.FORWARD);
    }
    inverse(spectrum, signal) {
        const work = this.work;
        checkFftBuffers(this.#memory, work, signal, spectrum);
        this.#impl.transformOrdered(spectrum.ptr, signal.ptr, work.ptr, Direction.BACKWARD);
    }
    forwardToInternalLayout(signal, spectrum) {
        const work = this.work;
        checkFftBuffers(this.#memory, work, signal, spectrum);
        this.#impl.transform(signal.ptr, spectrum.ptr, work.ptr, Direction.FORWARD);
    }
    inverseFromInternalLayout(spectrum, signal) {
        const work = this.work;
        checkFftBuffers(this.#memory, work, signal, spectrum);
        this.#impl.transform(spectrum.ptr, signal.ptr, work.ptr, Direction.BACKWARD);
    }
    reorder(spectrum, output) {
        const memory = this.#memory;
        if (memory !== spectrum._memory || memory !== output._memory) {
            throw mismatchedInstanceError();
        }
        const minN = (this.work.type === ValueType.COMPLEX ? 2 : 1) * this.n;
        if (spectrum.array.length < minN) {
            throw new Error("spectrum buffer too small");
        }
        if (output.array.length < minN) {
            throw new Error("output buffer too small");
        }
        this.#impl.zreorder(spectrum.ptr, output.ptr, Direction.FORWARD);
    }
    convolve(a, b, ab, scaling) {
        checkConvolveBuffers(this.#memory, this.work, a, b, ab);
        this.#impl.zconvolveNoAccu(a.ptr, b.ptr, ab.ptr, scaling);
    }
    convolveAccumulate(a, b, ab, scaling) {
        checkConvolveBuffers(this.#memory, this.work, a, b, ab);
        this.#impl.zconvolveAccu(a.ptr, b.ptr, ab.ptr, scaling);
    }
}
export class FFTLibrary {
    arrayType;
    type;
    #exports;
    constructor(arrayType, type, exports) {
        this.arrayType = arrayType;
        this.type = type;
        this.#exports = exports;
    }
    setup(n) {
        return new Setup(this.arrayType, this.type, n, this.#exports);
    }
    nearestValidSize(n, higher) {
        const ex = this.#exports;
        return (this.arrayType === Float32Array
            ? ex.pffft_nearest_transform_size
            : ex.pffftd_nearest_transform_size)(n, this.type, higher);
    }
    isValidSize(n) {
        const ex = this.#exports;
        return (this.arrayType === Float32Array
            ? ex.pffft_is_valid_size
            : ex.pffftd_is_valid_size)(n, this.type);
    }
    minFftSize() {
        const ex = this.#exports;
        return (this.arrayType === Float32Array
            ? ex.pffft_min_fft_size
            : ex.pffftd_min_fft_size)(this.type);
    }
}
export async function fetchWasm(url, importObject) {
    const { instance } = await WebAssembly.instantiateStreaming(fetch(url), importObject);
    const exports = instance.exports;
    exports._initialize();
    return exports;
}
const wasmConfigs = {
    f32: { path: "./pffft-ng-f32.wasm", exports: null },
    f64: { path: "./pffft-ng-f64.wasm", exports: null },
};
async function loadForConfig(config, importObject) {
    if (!config.exports) {
        const url = new URL(config.path, import.meta.url);
        config.exports = await fetchWasm(url, importObject);
    }
    return config.exports;
}
export async function loadLibrary(arrayType, type) {
    const config = arrayType === Float32Array ? wasmConfigs.f32 : wasmConfigs.f64;
    const exports = await loadForConfig(config);
    return new FFTLibrary(arrayType, type, exports);
}
//# sourceMappingURL=index.js.map